/*
 -------------------------------------------------------------------------
 Copyright (c) 2002-2006, XIMETA, Inc., FREMONT, CA, USA.
 All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary 
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright 
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products 
      built using this software without specific written permission. 
      
 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.
 
 DISCLAIMER

 This software is provided 'as is' with no explcit or implied warranties
 in respect of any properties, including, but not limited to, correctness 
 and fitness for purpose.
 -------------------------------------------------------------------------
*/
#include <linux/version.h> // LINUX_VERSION_CODE,KERNEL_VERSION
#include <linux/kernel.h> // container_of
#include <linux/module.h> // module_put for 2.6, MOD_INC/DEC for 2.4, EXPORT_SYMBOL
#include <linux/list.h> // list_node
#include <linux/notifier.h> // struct notifier_block
#include <linux/netdevice.h> // struct netdevice, register_netdevice_notifier, unregister_netdevice_notifier
#include <linux/spinlock.h> // spinlock
#include <linux/net.h> // sock_create
#include <linux/string.h> // memset
#ifdef NDAS_SNAP
#include <net/psnap.h> // register_snap_client
#include <net/p8022.h> // register_8022_client/unregister_8022_client
#endif
#include <net/datalink.h> // struct datalink_proto
#include <net/sock.h> // sock_no_socketpair, sock_no_setsockopt, sock_no_getsockopt, struct sock

#include <sal/net.h>
#include <sal/debug.h>
#include <sal/sync.h>
#include <sal/types.h>
#include <sal/mem.h>
#include <sal/thread.h>
//#include "lpx/lpx.h"
#include "linux.h"

#include "net_util.h"

//#define ENABLE_PACKET_DROP_TEST 1
#ifdef ENABLE_PACKET_DROP_TEST
static volatile int v_rx_packet_count;
static volatile int v_tx_packet_count;
#endif

#define PF_NDAS 29
#define ETH_P_NDAS    0x88ad

/**
 * From 2.6.9, skb->mac.ethernet is removed.
 * So we added ETH_DEST macro
 */
static inline char* ETH_DEST(struct sk_buff* skb) {
    return ((struct ethhdr *)skb->mac.raw)->h_dest;
}

#ifdef NDAS_SNAP
static unsigned char _8022_type = 0xD0;
static unsigned char _snap_id[5] = { 0x0, 0x0, 0x0, 0x88, 0xad };
#endif

struct proto_ops ndas_proto_ops;

unsigned long sal_net_recv_count=0;
unsigned long sal_net_drop_count=0;

static struct net_proto_family lpx_family;
/**
 * the network facility for the driver
 */
struct _network_facility v_net;

EXPORT_SYMBOL(sal_net_recv_count);
EXPORT_SYMBOL(sal_net_drop_count);
    
sal_net_buff init_net_buff(sal_buf_seg* bufs, struct sk_buff* skb) 
{
    /* link layer header */
    dbgl_salnet(9,"ing skb=%p",skb);
    dbgl_salnet(8,"skb=%s",DEBUG_SK_BUFF(skb));
    bufs->mac_header.ptr = skb->mac.raw;
    bufs->mac_header.len = sizeof(struct ethhdr);// skb->len - skb->data_len; 
    /* data */
    bufs->segs[0].ptr = skb->data;
    bufs->segs[0].len = skb->len;
    bufs->total_len = bufs->mac_header.len + bufs->segs[0].len;
    bufs->size = 1;    

    dbgl_salnet(8,"bufs->segs[0].ptr=%p",bufs->segs[0].ptr);
    return (sal_net_buff) skb;
}


/* wrapped packet receiver */
static inline int sal_ether_recv(struct sk_buff *skb, _network_device_t* pnd)
{
    sal_buf_seg bufs;
    sal_net_buff nbuf;

#ifdef ENABLE_PACKET_DROP_TEST
    v_rx_packet_count++;
    if (v_rx_packet_count%1000==0) {
        printk("(RXD)");
        goto drop;
    }
#endif


    if ( !pnd->rx_handler )
        goto drop; 
    
#ifdef NDAS_DEBUG_DROP_BROADCAST        
    // check if the packet is broadcasted
    if(memcmp(BROADCAST_NODE, ETH_DEST(skb), IFHWADDRLEN) == 0) {
        /* For debug purpose */
        goto drop; 
    }
#endif
    // check if the packet is for this device.
//    if(memcmp(pnd->mac, ETH_DEST(skb), IFHWADDRLEN) == 0)
//        goto forward;
    
    dbgl_salnet(9, "skb=%s",DEBUG_SK_BUFF(skb));    
    dbgl_salnet(6, "ing");
    nbuf = init_net_buff(&bufs, skb);

    (*pnd->rx_handler)((sal_netdev_desc)pnd,nbuf,&bufs);
    return 0; // nobody cares for the return value
drop:
    sal_net_drop_count++;
    kfree_skb(skb);
    return -1; // nobody cares for the return value
}
#ifdef NDAS_SNAP
/*
 * If the perfomance on snap/802.2 is important, optimize this. 
 * (hashtable or something) or (Accept the packet for any network device)
 */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)) 
int _802_rx_handler(struct sk_buff *buf, struct net_device *dev, struct packet_type *type) 
#else
int _802_rx_handler(struct sk_buff *buf, struct net_device *dev, struct packet_type *type, struct net_device * pdev) 
#endif
{
    unsigned long flags;

    _network_device_t* pos;
    struct list_head* i;
    sal_net_recv_count++;
    dbgl_salnet(5,"ing");
    sal_mutex_take(v_net.listlock, &flags);
    
    if ( list_empty(&v_net.listhead) ) 
        goto finally;
    
    list_for_each(i, &v_net.listhead)
    {
        pos = list_entry(i, _network_device_t, node);
        if (pos->ptype.dev == dev) {
            sal_mutex_give(v_net.listlock, &flags);
            if ( pos->down ) {
                kfree_skb(buf);
                sal_net_drop_count++;
                return 0;
            }
            return sal_ether_recv(buf,pos);
        }
    }
    
finally:     // drop the packet as the sal_netdev_register_proto is not called with this network device.
    sal_mutex_give(v_net.listlock, &flags);
    kfree_skb(buf);
    sal_net_drop_count++;
    return -1;
}
#endif /* NDAS_SNAP */
/* lpx packet receiver */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14))
int rx_handler(struct sk_buff *buf, struct net_device *dev, struct packet_type *type)
{
    _network_device_t* pnd;
    dbgl_salnet(9,"ing");
    sal_net_recv_count++;
    pnd = list_entry(type, _network_device_t, ptype);
    if ( pnd->down ) {
        kfree_skb(buf);
        sal_net_drop_count++;
        return 0;
    }
    
    return sal_ether_recv(buf,pnd);
}
#else
int rx_handler(struct sk_buff *buf, struct net_device *dev, struct packet_type *type, struct net_device *pdev)
{
    _network_device_t* pnd;
    dbgl_salnet(9,"ing");
    sal_net_recv_count++;
    pnd = list_entry(type, _network_device_t, ptype);
    if ( pnd->down ) {
        kfree_skb(buf);
        sal_net_drop_count++;
        return 0;
    }
    
    return sal_ether_recv(buf,pnd);
}
#endif

/**
 * _free_sal_next
 * 
 */
static void _free_device(_network_device_t* device)
{
    unsigned long flags;

    sal_mutex_take(v_net.listlock, &flags);
    list_del(&device->node); /* Unlink link */
    sal_mutex_give(v_net.listlock, &flags);
    
    sal_free(device); /* free */
}
#if LINUX_VERSION_SK_ALLOC_NEED_PROTO
static struct proto ndas_proto = {
	.name	  = "NDAS",
	.owner	  = THIS_MODULE,
	.obj_size = sizeof(struct sock),
};
#endif
static int    _sock_create(struct socket *sock, int protocol) 
{
    struct sock* sk;

    sk = SK_ALLOC(PF_NDAS, GFP_ATOMIC, 1, &ndas_proto); /* or sk->sk_slab*/
    if( !sk ) {
        dbgl_salnet(1,"fail to create sock");
        return -1;
    }

    sock->ops = &ndas_proto_ops;
    sock_init_data(sock, sk);
    sk->SK_DESTRUCT    = NULL;
    sk->SK_ALLOCATION = GFP_ATOMIC;
    return 0;
}
NDAS_CALL ndas_error_t (*sal_net_restart)(void) = NULL;

NDAS_SAL_API void sal_net_register_restart_func(NDAS_CALL ndas_error_t (*func)(void)) 
{
    sal_net_restart = func;
}
EXPORT_SYMBOL(sal_net_register_restart_func);

static int v_notifiee(struct notifier_block *self, unsigned long event, void *ptr)
{
    _network_device_t *ndev;
    struct net_device *dev = (struct net_device *) ptr;
    if ( !dev ) {
        dbgl_salnet(1, "ing dev=%p event=%ld", dev, event);
        return NOTIFY_DONE;
    }
    dbgl_salnet(1, "ing dev=%p(%s) event=%ld", dev, dev->name, event);
    if ( event == NETDEV_UP ) {
        ndev = FIND_NETWORK_DEVICE_STRUCT(dev->name);
        if ( ndev ) ndev->down = 0;

        if ( sal_net_restart )
            sal_net_restart();
        
        dbgl_salnet(1, "ed");
        return NOTIFY_DONE;
    }
    if ( event != NETDEV_DOWN )
        return NOTIFY_DONE;
    ndev = FIND_NETWORK_DEVICE_STRUCT(dev->name);
    if ( ndev ) {
        ndev->down = 1;
    }
    
    return NOTIFY_DONE;
}
static int ndas_sock_no_release(struct socket *sock) { return 0; }
/**
 * sal_net_init - initialize the SAL network layer
 * register rx handler to ether packet, 8022, snap and 
 * register net device event notifier. 
 */
NDAS_SAL_API int sal_net_init(void)
{
    dbgl_salnet(1,"ing");

    v_net.listlock = sal_mutex_create("sal_net_list");
    sal_assert(v_net.listlock != SAL_INVALID_MUTEX);
    lpx_family.family = PF_NDAS;
    lpx_family.create = _sock_create;
    v_net.notifier = NULL;
    INIT_LIST_HEAD(&v_net.listhead);
#ifdef NDAS_SNAP
    v_net.p8022 = NULL;
    v_net.psnap = NULL;

//    request_module("psnap");
//    request_module("p8022");
    // TODO check the drivers(llc, 8022, snap) loaded 
    if (!v_net.p8022)
        v_net.p8022 = register_8022_client(_8022_type, _802_rx_handler);
    
    if (!v_net.p8022)
        sal_error_print("ndas: Unable to register with 802.2\n");
                                                                            
    if (!v_net.psnap)
        v_net.psnap = register_snap_client(_snap_id, _802_rx_handler);
    
    if (!v_net.psnap)
        sal_error_print("ndas: Unable to register with SNAP\n");
#endif /* NDAS_SNAP */

    ndas_proto_ops.family = PF_NDAS;
    ndas_proto_ops.release = ndas_sock_no_release;
    ndas_proto_ops.bind = sock_no_bind;
    ndas_proto_ops.connect = sock_no_connect;
    ndas_proto_ops.socketpair = sock_no_socketpair;
    ndas_proto_ops.accept = sock_no_accept;
    ndas_proto_ops.getname = sock_no_getname;
    ndas_proto_ops.poll = datagram_poll;  /* this does seqlpx too */
    ndas_proto_ops.ioctl = sock_no_ioctl;
    ndas_proto_ops.listen = sock_no_listen;
    ndas_proto_ops.shutdown = sock_no_shutdown;
    ndas_proto_ops.setsockopt = sock_no_setsockopt;
    ndas_proto_ops.getsockopt = sock_no_getsockopt;
    ndas_proto_ops.sendmsg = sock_no_sendmsg;
    ndas_proto_ops.recvmsg = sock_no_recvmsg;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0))
    ndas_proto_ops.owner = THIS_MODULE;
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0))
    ndas_proto_ops.mmap = sock_no_mmap;
    ndas_proto_ops.sendpage = sock_no_sendpage;
#else    
    ndas_proto_ops.dup = sock_no_dup;
    ndas_proto_ops.fcntl = sock_no_fcntl;
#endif    

    if ( sock_register(&lpx_family) !=0 ) {
        printk("ndas: fail to register ndas network protocol family\n"); 
        sal_net_cleanup();
        return -1;
    }

    v_net.notifier = kmalloc(sizeof(struct notifier_block), GFP_ATOMIC);
    memset(v_net.notifier, 0, sizeof(struct notifier_block));
    v_net.notifier->notifier_call = v_notifiee;
    v_net.notifier->next = NULL;
    v_net.notifier->priority = 0;
    
    register_netdevice_notifier(v_net.notifier);

    dbgl_salnet(1,"ed");
    return 0;
}
EXPORT_SYMBOL(sal_net_init);

NDAS_SAL_API void sal_net_cleanup(void)
{
    if ( v_net.notifier ) { 
        unregister_netdevice_notifier(v_net.notifier);
        kfree(v_net.notifier);
        v_net.notifier = NULL;
    }
    sock_unregister(PF_NDAS);
#ifdef NDAS_SNAP
    if (v_net.p8022) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0))
        unregister_8022_client(v_net.p8022);
#else
        unregister_8022_client(_8022_type);
#endif 
    }
    v_net.p8022 = NULL;
    if (v_net.psnap) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0))
        unregister_snap_client(v_net.psnap);
#else
        unregister_snap_client(_snap_id);
#endif
    }
    v_net.psnap = NULL;
#endif /* NDAS_SNAP */
    sal_mutex_destroy(v_net.listlock);
}
EXPORT_SYMBOL(sal_net_cleanup);

/**
 * sal_netdev_open - open the network device by the given name
 * @devname - the name of the network device to open
 */
NDAS_SAL_API sal_netdev_desc sal_netdev_open(const char* devname)
{
    unsigned long flags;
    int ret;
    _network_device_t* psn = NULL;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
    struct device     *dev;
#else
    struct net_device     *dev;
#endif

    dbgl_salnet(3, "ing name=%s", devname);
    if ( strcmp(devname, "lo") == 0 )
        return SAL_INVALID_NET_DESC;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
    dev = dev_get(devname);
#else
    dev = dev_get_by_name(devname);
#endif
    dbgl_salnet(3, "..");
    if (dev == NULL) {
        dbgl_salnet(1, "no such device %s:",devname);
        return SAL_INVALID_NET_DESC;
    }
    dbgl_salnet(3, "..");
    if ( !list_empty(&v_net.listhead) && FIND_NET_DEVICE_STRUCT(devname) )
    {
        dbgl_salnet(1, "already registered device %s:",devname);
        return SAL_INVALID_NET_DESC; 
    }
    dbgl_salnet(3, "..");
    /* Now everything OK, 
        create place holder for new net interface */
        
    psn = (_network_device_t*) sal_malloc(sizeof(_network_device_t));
    if ( !psn ) {
        dbgl_salnet(1, "out of memory %s:",devname);
        return SAL_INVALID_NET_DESC;
    }
    memset(psn, 0, sizeof(_network_device_t));
    
    strncpy(psn->devname, devname, IFNAMSIZ); // TODO: strncpy 
    psn->ptype.dev = dev;
    psn->is_bound = FALSE;

    ret = sal_netdev_get_address((sal_netdev_desc)psn, psn->mac);
    if (ret!=0) {
        goto errout;
    }

    /* Link to net list */
    sal_mutex_take(v_net.listlock, &flags);
    
    list_add(&psn->node, &v_net.listhead); 
    sal_mutex_give(v_net.listlock, &flags);

    dbgl_salnet(5, "sal_netdev_opened : %s", devname);
    return (sal_netdev_desc)psn;
    
errout:
    dbgl_salnet(1, "sal_netdev_open : can\'t open %s", devname);
    sal_free(psn); // psn is not NULL
    return 0;
    
}
EXPORT_SYMBOL(sal_netdev_open);

NDAS_SAL_API int sal_netdev_close(sal_netdev_desc nd)
{
    _network_device_t* dev = GET_NETWORK_DEVICE_T(nd);
    dbgl_salnet(1, "sal_netdev_close starting");
    if (!dev) {
        return -1;    
    }
    dev->down = TRUE;
    _free_device(dev);
    dbgl_salnet(1, "sal_netdev_close succeeded");
    return 0;
}
EXPORT_SYMBOL(sal_netdev_close);

NDAS_SAL_API int sal_netdev_get_mtu(sal_netdev_desc nd)
{
    _network_device_t* dev = GET_NETWORK_DEVICE_T(nd);
    return GET_NET_DEVICE_STRUCT(dev)->mtu;
}
EXPORT_SYMBOL(sal_netdev_get_mtu);

/* 
 allocate sturct sk_buff
 */
NDAS_SAL_API sal_net_buff sal_net_alloc_platform_buf(sal_sock nd, xuint32 len, sal_buf_seg* bufs)
{
    struct sk_buff *skb;
    struct sock* sk = (struct sock*) nd;
//  unsigned long flags;
//  spinlock_t lock = SPIN_LOCK_UNLOCKED;
    int non_blocking_call = 0;
    int err;
    dbgl_salnet(3, "nd=%p len=%d bufs=%p", nd, len, bufs);
    dbgl_salnet(5, "sal_net_alloc_platform_buf: sk=%s, len=%d, bufs=%p", 
                    DEBUG_LINUX_SOCK(sk), len, bufs);
    /* check for disbling preemption or IRQs*/
//spin_lock_irqsave(&lock,flags); 
/*    if (atomic_read(&sk->wmem_alloc)>64*1024) {
    printk("(%d)\n",atomic_read(&sk->wmem_alloc));
    }*/
    skb = sock_alloc_send_skb(sk, len ,non_blocking_call , &err); 
//spin_unlock_irqrestore(&lock,flags);

    if ( !skb ) {
    printk("Failed to allocate skb len=%d\n", len);
    return NULL;
    }
    bufs->mac_header.ptr = skb->mac.raw = skb->data;
    skb_reserve(skb, SAL_ETHER_HEADER_SIZE);
    // TODO: size should be different for 802.2 snap , and added 2 for alignment
    bufs->segs[0].ptr = 
    skb->h.raw = 
    skb->nh.raw = skb_put(skb,len - SAL_ETHER_HEADER_SIZE);
    bufs->segs[0].len = len - SAL_ETHER_HEADER_SIZE;
    bufs->size = 1;
    bufs->total_len = len;

    dbgl_salnet(9, "ed skb=%s",_DEBUG_SK_BUFF_P(skb));
    return (sal_net_buff) skb;
}
EXPORT_SYMBOL(sal_net_alloc_platform_buf);

NDAS_SAL_API int sal_net_free_platform_buf(sal_net_buff buf)
{
    dbgl_salnet(9, "ing"); 
    kfree_skb((struct sk_buff *) buf);
    dbgl_salnet(9, "ed");
    return 0;
}
EXPORT_SYMBOL(sal_net_free_platform_buf);

NDAS_SAL_API sal_net_buff sal_net_clone_platform_buf(sal_net_buff buf)
{
    struct sk_buff* skb;
    skb = skb_clone((struct sk_buff*)buf, GFP_ATOMIC);
    if (skb==NULL) {
        printk("Failed to clone skb\n");
    } else {
/*	printk("skb_clone: skb=%p, mac=%p, data=%p -> skb=%p, mac=%p, data=%p\n",
		(char*) buf, ((struct sk_buff*)buf)->mac.raw, ((struct sk_buff*)buf)->data, 
		(char*) skb, skb->mac.raw, skb->data);*/
    }
    return (sal_net_buff) skb;
}

EXPORT_SYMBOL(sal_net_clone_platform_buf);

NDAS_SAL_API sal_net_buff sal_net_copy_platform_buf(sal_net_buff buf, xuint32 len, sal_buf_seg* bufs)
{
    struct sk_buff* skb;
    skb = skb_copy((struct sk_buff*)buf, GFP_ATOMIC);
    if (skb==NULL) {
        printk("Failed to copy skb\n");
	return NULL;
    }

    bufs->mac_header.ptr = skb->mac.raw;
    bufs->segs[0].ptr = skb->data;
    bufs->segs[0].len = len - SAL_ETHER_HEADER_SIZE;
    bufs->size = 1;
    bufs->total_len = len;

    return (sal_net_buff) skb;
}

EXPORT_SYMBOL(sal_net_copy_platform_buf);

/**
 * sal_netdev_register_proto
 */
NDAS_SAL_API int sal_netdev_register_proto(sal_netdev_desc nd, xuint16 protocolid, sal_net_rx_proc rxhandler)
{
    _network_device_t* pnd = GET_NETWORK_DEVICE_T(nd);
    dbgl_salnet(5, "ing");

    if (!pnd) {
        return -1;    
    }
    pnd->ptype.type = protocolid;
    pnd->ptype.func = rx_handler;
    dev_add_pack(&pnd->ptype); /* register the package type */
    pnd->rx_handler = rxhandler;
    dbgl_salnet(5, "ed");
    return 0;
}
EXPORT_SYMBOL(sal_netdev_register_proto);

NDAS_SAL_API int sal_netdev_unregister_proto(sal_netdev_desc nd, xuint16 protocolid) 
{
    _network_device_t* pnd = GET_NETWORK_DEVICE_T(nd);
    dbgl_salnet(1, "ing");
    if (!pnd) {
        return -1;    
    }
    sal_assert(protocolid == pnd->ptype.type);
    dev_remove_pack(&pnd->ptype);
    dbgl_salnet(1, "ed");
    return 0;
}
EXPORT_SYMBOL(sal_netdev_unregister_proto);
/**
 * sal_netdev_set_address 
 * Set the virtual mac address of the network device
 * @param nd the network descriptor
 * @param mac the virtual mac address to be set
 */
NDAS_SAL_API int sal_netdev_set_address(sal_netdev_desc nd, unsigned char* mac)
{
    _network_device_t* sn = GET_NETWORK_DEVICE_T(nd);
    if (!sn) {
        return -1;    
    }
    
    sn->use_virtual_mac = 1;
    memcpy(sn->virtual_mac, mac, IFHWADDRLEN);
    return 0;
}
EXPORT_SYMBOL(sal_netdev_set_address);

/**
 * sal_netdev_get_address -
 * Get the mac address of the network device
 * @param nd the net descriptor
 * @param 6 bytes mac address found (out parameter)
 * @return 0 for OK, negative value for error 
 */
NDAS_SAL_API int sal_netdev_get_address(sal_netdev_desc nd, unsigned char* mac)
{
    _network_device_t* pnd = GET_NETWORK_DEVICE_T(nd);
    if (!pnd) {
        return -1;    
    }
    
    if (pnd->use_virtual_mac) {
        memcpy(mac, pnd->virtual_mac, IFHWADDRLEN);
        return 0;
    }
    /* get hardware address */
    if ( pnd->ptype.dev->addr_len != IFHWADDRLEN ) {
        dbgl_salnet(1, "the size of mac address is %d bytes:", pnd->ptype.dev->addr_len);
        return -1;
    }
    memcpy(mac, pnd->ptype.dev->dev_addr, pnd->ptype.dev->addr_len);

    return 0;
}
EXPORT_SYMBOL(sal_netdev_get_address);
/**
 * sal_ether_send
 */
NDAS_SAL_API int sal_ether_send(sal_netdev_desc nd, sal_net_buff buf, xuint32 len)
{
    struct sk_buff* skb = (struct sk_buff*)buf;
    _network_device_t* pnd;

#ifdef ENABLE_PACKET_DROP_TEST
    v_tx_packet_count++;
    if (v_tx_packet_count%1000==0) {
        printk("TXD");
        sal_net_free_platform_buf(buf);
        return -3;
    }
#endif

    dbgl_salnet(6, "len=%d", len);
    pnd = GET_NETWORK_DEVICE_T(nd);
    if (!pnd || pnd->down ) {
        sal_net_free_platform_buf(buf);
        dbgl_salnet(1, "len=%d", len);
        return -3;
    }
//    sal_debug_print(1, "sal_ether_send to %s, %d bytes", pnd->devname, len);


    if(pnd->ptype.dev->flags & IFF_LOOPBACK) {
        debug_sal_net_packet(1, "send to local loop is not supported yet");
    }
    skb->dev = pnd->ptype.dev;
    skb->protocol = htons(ETH_P_NDAS);
    
    if(skb->dev->hard_header) {
        char dest_node[IFHWADDRLEN];
        if(memcmp(BROADCAST_NODE, ETH_DEST(skb), IFHWADDRLEN) == 0) { 
            memcpy(dest_node, skb->dev->broadcast, IFHWADDRLEN);    
        } else {
            memcpy(dest_node, ETH_DEST(skb), IFHWADDRLEN);
        }

        skb->dev->hard_header(skb, skb->dev, ETH_P_NDAS, dest_node, NULL, skb->len);
    }

    debug_sal_net_packet(7, "data=%p size=%d", skb->data,skb->len);
    debug_sal_net_packet(8, "%s", SAL_DEBUG_HEXDUMP(skb->data,skb->len));

    dev_queue_xmit(skb);
    debug_sal_net_packet(6, "ed");
    return len;
}
EXPORT_SYMBOL(sal_ether_send);

/**
 * create the system socket for internal usage of LPX socket.
 * system socket 'struct sock' should be create for allocating sk_buff.
 * even the lpx socket is created directly inside the driver.
 */
NDAS_SAL_API sal_sock sal_sock_create(int type, int protocol, int mts)
{
    struct socket* sock;
    int err;
    //MOD_INC_USE_COUNT;  TODO: the owner of this count should be ndas_core

    dbgl_salnet(5, "type=%d",type);

    err = SOCK_CREATE(PF_NDAS, type, 0,&sock);

    if ( err ) {
        dbgl_salnet(1,"fail to create socket");
        //MOD_DEC_USE_COUNT;
        return NULL;
    }
    sal_assert(sock->sk);

#if LINUX_VERSION_25_ABOVE  
    sock->sk->sk_sndbuf = mts + mts/2;
#else
    sock->sk->sndbuf = mts + mts/2; // Varies on lpx configuration. Increase this if sock_alloc_send_skb fails
#endif
    dbgl_salnet(5, "created socket=%p",sock);
    return (sal_sock) sock->sk;
}
EXPORT_SYMBOL(sal_sock_create);

NDAS_SAL_API void sal_sock_destroy(sal_sock sk0) {
    struct sock* sk = (struct sock*) sk0; 
    struct socket* sock;
    dbgl_salnet(5, "sal_sock=%s",DEBUG_LINUX_SOCK(sk));
    
    sal_assert(sk0 != NULL); // fix me
    
    sock = sk->SK_SOCKET;
    if ( sk ) sk_free(sk);
    if ( sock ) sock_release(sock);
    
    //MOD_DEC_USE_COUNT; 
}
EXPORT_SYMBOL(sal_sock_destroy);
