/*
 -------------------------------------------------------------------------
 Copyright (c) 2002-2006, XIMETA, Inc., FREMONT, CA, USA.
 All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary 
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright 
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products 
      built using this software without specific written permission. 
      
 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.
 
 DISCLAIMER

 This software is provided 'as is' with no explcit or implied warranties
 in respect of any properties, including, but not limited to, correctness 
 and fitness for purpose.
 -------------------------------------------------------------------------
*/
#ifndef __SAL_NET_UTIL_H_
#define __SAL_NET_UTIL_H_

#include <linux/netdevice.h> // struct netdevice, register_netdevice_notifier, unregister_netdevice_notifier
#include "sal/net.h" // SAL_NET_MAX_DEV_NAME_LENGTH
#include "sal/debug.h"
#include "sal/sync.h" // sal_semaphore

#define MAX_PACKET_LENGTH 1600

#define BROADCAST_NODE    "\xFF\xFF\xFF\xFF\xFF\xFF"

#define GET_LPX_FD(sock) ((int)sock->SK_USER_DATA)
#define SET_LPX_FD(sock,fd) (sock->SK_USER_DATA = (void*) fd)

#ifdef DEBUG
    #ifndef DEBUG_LEVEL_SAL_NET
        #ifdef DEBUG_LEVEL
            #define DEBUG_LEVEL_SAL_NET DEBUG_LEVEL
        #else
            #define DEBUG_LEVEL_SAL_NET 0
        #endif
    #endif
    #ifndef DEBUG_LEVEL_SAL_NET_PACKET
        #ifdef DEBUG_LEVEL
            #define DEBUG_LEVEL_SAL_NET_PACKET DEBUG_LEVEL
        #else
            #define DEBUG_LEVEL_SAL_NET_PACKET 0
        #endif
    #endif    
#endif

#ifdef DEBUG
#define dbgl_salnet(l, x...) do { \
    if ( l <= DEBUG_LEVEL_SAL_NET ) {\
        sal_debug_print("SN|%d|%s|",l,__FUNCTION__);\
        sal_debug_println(x); \
    }\
} while(0)
#define debug_sal_net_packet(l, x...) do { \
    if ( l <= DEBUG_LEVEL_SAL_NET_PACKET ) {\
        sal_debug_print("SNP|%d|%s|",l,__FUNCTION__);\
        sal_debug_println(x); \
    }\
} while(0)
#else
#define dbgl_salnet(l, x...) do {} while(0)
#define debug_sal_net_packet(l, x...) do {} while(0)
#endif

#define DEBUG_SK_BUFF(skb) \
({\
    static char __buf__[1024];\
    if (!skb)\
        snprintf(__buf__,sizeof(__buf__),"NULL");\
    else\
        snprintf(__buf__,sizeof(__buf__),\
            "struct sk_buff(%p)={mac(%p)=%s,"\
            "h(%p),data(%dbytes,%p)=%s,tail=%p,"\
            "sock(%p),data_len=%d,head=%p,end=%p}",\
            skb,\
            skb->mac.raw,\
            SAL_DEBUG_HEXDUMP_S(skb->mac.raw, IFHWADDRLEN),\
            skb->h.raw,\
            skb->len,\
            skb->data,\
            SAL_DEBUG_HEXDUMP(skb->data,skb->len),\
            skb->tail,\
            skb->sk,\
            skb->data_len,\
            skb->head,\
            skb->end\
        );\
    (const char*) __buf__;\
})

#define _DEBUG_SK_BUFF_P(skb) \
({\
    static char __buf__[1024];\
    if (!skb)\
        snprintf(__buf__,sizeof(__buf__),"NULL");\
    else\
        snprintf(__buf__,sizeof(__buf__),\
            "struct sk_buff(%p)={mac(%p),"\
            "nh(%p),h(%p),data(%dbytes,%p),tail=%p,"\
            "sock(%p),data_len=%d,head=%p,end=%p}",\
            skb,\
            skb->mac.raw,\
            skb->nh.raw, \
            skb->h.raw,\
            skb->len,\
            skb->data,\
            skb->tail,\
            skb->sk,\
            skb->data_len,\
            skb->head,\
            skb->end\
        );\
    (const char*) __buf__;\
})
    
#define DEBUG_LINUX_SOCK(sk) \
({\
    static char __buf__[1024];\
    if (!sk)\
        snprintf(__buf__,sizeof(__buf__),"NULL");\
    else\
        snprintf(__buf__,sizeof(__buf__),\
            "struct sock(%p)={socket=%p,"\
            "allocation=0x%x}",\
            sk, sk->SK_SOCKET, \
            sk->SK_ALLOCATION\
        );\
    (const char*) __buf__;\
})
    

extern struct proto_ops ndas_proto_ops;

/**
 * _network_device_t
 * @node - circle-linked list node.
 * @devname - the name of network device
 * @mac - the 6 bytes mac address
 * @ptype - the packet type structure 
 * @ptype.dev - the network device
 * @rx_handler - forward call when ptype.rx_handler is called.
 */
typedef struct _network_device {
    struct list_head node;
    struct packet_type ptype;
    char    devname[IFNAMSIZ];
    char    mac[IFHWADDRLEN]; /* MAC address of this interface */
    sal_net_rx_proc rx_handler;
    int is_bound;    /* TRUE if interface is bound */
    int down;
    int use_virtual_mac; /* Used to change my MAC address to fake MAC. Used for netdisk emulation.*/
    char virtual_mac[IFHWADDRLEN];
} _network_device_t;

/**
 * the network facility for the driver
 */
struct _network_facility {
    struct list_head listhead; // list node
    sal_mutex listlock;
    struct datalink_proto *p8022;
    struct datalink_proto *psnap;
    struct notifier_block *notifier;
};

extern struct _network_facility v_net;
/*
 * Utilities for struct_network_device
 */
static inline struct net_device* GET_NET_DEVICE_STRUCT(struct _network_device* _nd) {
    return _nd->ptype.dev;
}

static inline _network_device_t* FIND_NETWORK_DEVICE_STRUCT(char* devname) 
{
    _network_device_t* pos;
    struct list_head* i;
    unsigned long flags;

    sal_mutex_take(v_net.listlock, &flags);

    if ( list_empty(&v_net.listhead) ) 
        goto finally;
    
    list_for_each(i, &v_net.listhead)
    {
        pos = list_entry(i, _network_device_t, node);
        if ( !strncmp(pos->devname,devname,IFNAMSIZ) ) 
        {
            sal_mutex_give(v_net.listlock, &flags);
            return pos;
        }
    }

finally:    
    sal_mutex_give(v_net.listlock, &flags);
    return NULL;
}

static inline struct net_device* FIND_NET_DEVICE_STRUCT(const char* devname) 
{
    _network_device_t* pos;
    struct list_head* i;
    unsigned long flags;
    sal_mutex_take(v_net.listlock, &flags);
    if ( list_empty(&v_net.listhead) ) 
        goto finally;
    list_for_each(i, &v_net.listhead)
    {
        pos = list_entry(i, _network_device_t, node);
        unlikely(pos == NULL);
        if ( !strncmp(pos->devname,devname,IFNAMSIZ) ) 
        {
            sal_mutex_give(v_net.listlock, &flags);    
            return pos->ptype.dev;
        }
    }
finally:    
    sal_mutex_give(v_net.listlock, &flags);    
    return NULL;
}

static inline _network_device_t* GET_NETWORK_DEVICE_T(sal_netdev_desc nd)
{
    return (_network_device_t*) nd;
}

extern sal_net_buff init_net_buff(sal_buf_seg* bufs, struct sk_buff* skb);

#endif /* __SAL_NET_UTIL_H_ */

